!--------------------------------------------------------------------------
! SMARTCANTGO.H
! David Wagner
! Version 1
! 22-Mar-1995
! All rights given away.
! 
! Defines a routine that can be used for the cant_go property of rooms to
! print the exits, instead of just "You can't go that way."  For example,
! the following code
! 
!     Object  Crystal_Cave "Crystal Cave"
!      with   s_to Narrow_Passage,
!             cant_go #r$SmartCantGo,
! 
! will produce the message "The exit is south." if the player goes the
! wrong way.
! 
! This code can be stored in a file and included into a game with 'Include
! "SmartCantGo"' sometime after the parser is included.
! 
! NOTES (far too many of them):
! 
!  1. If the room is dark, it just prints "You can't go that way."
!     How could the player know where the exits are?
!
!  2. This routine ignores direction properties which point to concealed
!     doors or are strings.  Therefore this code won't work quite as
!     intended in the following room:
!
!       Object  Library "Library"
!        with   description "A small wood-panelled library.  An open \
!                   window to the west affords a stunning view of \
!                   the Tuscan coastline.",
!               w_to "Ouch! You discover that the ~window~ is really \
!                   in incredibly lifelike mural painted on the wall.",
!
!  3. This routine does include direction properties that are routines
!     on the assumption that the routine will return something sensible.
!     If your routine returns a concealed door, you will have to do
!     something different.
!
!  4. This routine describes all the exits (including doors) as "exits." 
!     One could imagine modifying the routine to list non-door exits
!     separately from doors.  However, a "door" in a game might actually
!     be a "gate" or a "bridge" or a "wall of flame" or any other kind
!     of obstacle, and there is no way for this routine to know how a
!     door should be described.
!
!  5. Unfortunately, version 5/9 of the library doesn't support default
!     properties which are routines (due to  a problem with RunRoutines()).  
|     In other words, you can't use ChangeDefault() for cant_go; you must 
|     say "cant_go #r$SmartCantGo" for each room where you want it.
!
!  6. This routine is adequate for the rare occasions when the player
!     types the wrong direction.  One might be tempted to use this
!     routine to simplify room descriptions, as in:
!
!       Object  Boring_Room "Boring Room"
!        with   description [;
!                   "A small room with white walls. ";
!                   SmartCantGo();
!               ];
!
!     Don't do this!  Part of the interest of a room is a good
!     description of the exits.
!
!     "To the west is a wooden door, above which is strange gothic
!     lettering."
!--------------------------------------------------------------------------

[ SmartCantGo room i dest desttype ndir idir;

    ! How would a player in darkness know where the exits are?
    if (location == thedark) "You can't go that way.";

    !  Find what room the player is in.
    room = location;
    while (parent(room) ~= 0) room = parent(room);

    ! Count the number of exits-- if a direction property is a string or
    ! a concealed door, don't count it; if it is a routine, count it.
    objectloop (i in compass) {
        dest = room.(i.door_dir);
        if (dest ~= 0) {
            desttype = ZRegion(dest);
            if (desttype ~= 3 && 
              (desttype == 2 || dest hasnt door || dest hasnt concealed))
                ndir++;
        }
    }

    ! There might not be a (visible) exit
    if (ndir == 0) "There are no exits.";
    print "You can only go ";

    ! Print the exits.  (Note that we rely on the fact that the full name
    ! of the direction is stored as the second item in the list of name
    ! properties for the direction objects; true for the standard
    ! directions in version 5/9 of the Inform library.)
    objectloop (i in compass) {
        dest = room.(i.door_dir);
        if (dest ~= 0) {
            desttype = ZRegion(dest);
            if (desttype ~= 3 &&
              (desttype == 2 || dest hasnt door || dest hasnt concealed)) {
                idir++;
                print_addr (i.&name)-->1;
                if (idir <= ndir-2) print ", ";
                if (idir == ndir-1) print " and ";
            }
        }
    }
    ".";
];
